from sklearn.model_selection import train_test_split

def split_dataset(X, y, sizes=[0.2], rand_state=42):
    if isinstance(sizes, float):
        sizes = [sizes]
    elif isinstance(sizes, int):
        sizes = [sizes]
    elif isinstance(sizes, list):
        pass
    else:
        sizes = list(sizes)
    
    results = []
    current_X, current_y = X, y
    
    remaining = 1.0
    for size in sizes[:-1]:
        adj_size = size / remaining
        X_train, X_split, y_train, y_split = train_test_split(
            current_X, current_y, 
            test_size=adj_size,
            random_state=rand_state
        )
        results.append((X_split, y_split))
        current_X, current_y = X_train, y_train
        remaining -= size
    
    results.append((current_X, current_y))
    
    return tuple(results)



import torch
from torch import nn
from sklearn.model_selection import train_test_split
import numpy as np
from gamma.optim import _optimizer
from gamma.loss import _loss_fn

def is_same_device(t, device):
    if isinstance(t, torch.Tensor):
        return t.device == device
    elif isinstance(t, np.ndarray):
        return device == torch.device('cpu') or device == 'cpu' or device is None
    elif isinstance(t, list):
        return all(is_same_device(item, device) for item in t)
    elif isinstance(t, nn.Module):
        params = list(t.parameters())
        return len(params) == 0 or params[0].device == device
    
    raise ValueError(f"Invalid type: {type(t)}")

def get_device(t):
    if isinstance(t, torch.Tensor):
        return t.device
    elif isinstance(t, np.ndarray):
        return torch.device('cpu')
    elif isinstance(t, list):
        return get_device(t[0])
    elif isinstance(t, nn.Module):
        params = list(t.parameters())
        return params[0].device if len(params) > 0 else torch.device('cpu')
    
    return torch.device('cpu')

def to_tensor(x):
    if isinstance(x, torch.Tensor):
        return x
    return torch.tensor(x)
class Trainer:
    def __init__(self, model, optimizer, loss_fn, after_train=None, lr=0.01, device=None):
        self.model = model
        self.optimizer = _optimizer(model, optimizer, lr)
        self.loss_fn   = _loss_fn(loss_fn)
        self.after_train = after_train
        self.device = device
        if self.device is None: # if device is not specified, use the device of the model
            self.device = get_device(self.model)

    def train(self, X, y, val_data=None, split_val=False, epochs=10, batch_size=32):
        if not is_same_device(self.model, self.device):
            self.model = self.model.to(self.device)
        
        X, y = to_tensor(X), to_tensor(y)

        has_val = True
        if split_val:
            X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=0.2, random_state=42)
        else:
            X_train, y_train = X, y
            if val_data is None:
                has_val = False
            else:
                X_val, y_val = val_data

        optimizer = self.optimizer

        history = {'train_loss': [], 'val_loss': []}

        for epoch in range(epochs):
            self.model.train()
            running_loss = 0.0

            for i in range(0, len(X_train), batch_size):
                X_batch = X_train[i:i+batch_size]
                y_batch = y_train[i:i+batch_size]
                
                if not is_same_device(X_batch, self.device):
                    X_batch = X_batch.to(self.device)
                if not is_same_device(y_batch, self.device):
                    y_batch = y_batch.to(self.device)

                optimizer.zero_grad()

                y_hat = self.model(X_batch)
                loss = self.loss_fn(y_hat, y_batch)

                loss.backward()
                optimizer.step()

                running_loss += loss.item() * len(X_batch)

            train_loss = running_loss / len(X_train)
            history['train_loss'].append(train_loss)

            if has_val:
                self.model.eval()
                val_loss = 0.0

                with torch.no_grad():
                    for i in range(0, len(X_val), batch_size):
                        X_batch = X_val[i:i+batch_size]
                        y_batch = y_val[i:i+batch_size]

                        outputs = self.model(X_batch)
                        loss = self.loss_fn(outputs, y_batch)

                        val_loss += loss.item() * len(X_batch)

                val_loss = val_loss / len(X_val)
                history['val_loss'].append(val_loss)

            print(f'Epoch {epoch+1}/{epochs} - train_loss: {train_loss:.4f} - val_loss: {val_loss:.4f}')

        return history
    
    def evaluate(self, X, y, batch_size=32):
        running_loss = 0.0
        self.model.eval()
        with torch.no_grad():
            for i in range(0, len(X), batch_size):
                X_batch = X[i:i+batch_size]
                y_batch = y[i:i+batch_size]
                loss = self.loss_fn(self.model(X_batch), y_batch)
                running_loss += loss.item() * len(X_batch)
        return running_loss / len(X)



import torch

def _str_to_optim(optm):
    optm = optm.lower()
    match optm:
        case 'adam':
            return torch.optim.Adam
        case 'sgd':
            return torch.optim.SGD
        case 'rmsprop':
            return torch.optim.RMSprop
        case 'adagrad':
            return torch.optim.Adagrad
        case 'adadelta':
            return torch.optim.Adadelta
        case 'adamw':
            return torch.optim.AdamW
        case _:
            raise ValueError(f"Unsupported optimizer name: {optm}")

def _optimizer(model, optimizer, lr=0.01):
    if optimizer is None:
        return torch.optim.Adam(model.parameters(), lr=lr)
    
    if isinstance(optimizer, torch.optim.Optimizer):
        optimizer_type = type(optimizer)
        param_groups = []
        for group in optimizer.param_groups:
            new_group = {k: v for k, v in group.items() if k != 'params'}
            new_group['params'] = model.parameters()
            param_groups.append(new_group)
        return optimizer_type(param_groups)
    
    if isinstance(optimizer, str):
        return _str_to_optim(optimizer)(model.parameters(), lr=lr)
    
    if isinstance(optimizer, dict):
        opt_type = optimizer.get('type', 'adam').lower()
        params = {k: v for k, v in optimizer.items() if k != 'type'}
        return _str_to_optim(opt_type)(model.parameters(), **params)

    else:
        raise ValueError(f"Invalid optimizer type: {type(optimizer)}")



import torch
import torch.nn as nn

class Lambda(nn.Module):
    def __init__(self, f):
        super().__init__()
        self.f = f

    def forward(self, x):
        return self.f(x)

class swap1(Lambda):
    def __init__(self):
        super().__init__(lambda x: x.swapaxes(1, 2))

class swap2(Lambda):
    def __init__(self):
        super().__init__(lambda x, y: (y, x))

class assert_shape(nn.Module):
    def __init__(self, shape):
        super().__init__()
        self.shape = shape

    def forward(self, x):
        assert x.shape == self.shape, f"Shape mismatch. Expected shape {self.shape}, got {x.shape}."
        return x

class assert_shape_len(Lambda):
    def __init__(self, shape_len):
        super().__init__(lambda x: x if len(x.shape) == shape_len else ValueError(f"Shape mismatch. Expected shape length {shape_len}, got {len(x.shape)}."))

class assert_(Lambda):
    def __init__(self, f, error_message="Assertion failed"):
        super().__init__(lambda x: x if f(x) else ValueError(error_message))

class print_shape(Lambda):
    def __init__(self):
        super().__init__(lambda x: print(x.shape))

class print_(Lambda):
    def __init__(self):
        super().__init__(lambda x: print(x))

class fork(nn.Module):
    def __init__(self, N=2):
        super().__init__()
        self.N = N

    def forward(self, x):
        return [x for _ in range(self.N)]

class parallel(nn.Module):
    def __init__(self, *modules):
        super().__init__()
        self.modules = nn.ModuleList(modules)

    def forward(self, x):
        return [module(x) for module in self.modules]


def seq(*blocks):
    return nn.Sequential(*blocks)

def _ruby_unbox(block):
    if len(block) == 1:
        return block[0]
    return [ruby_pipeline_element(b) for b in block]

def ruby_pipeline_element(block):
    if isinstance(block, tuple):
        block = seq(*_ruby_unbox(block))
    elif isinstance(block, list):
        block = parallel(*_ruby_unbox(block))
    return block

def pipeline(*blocks):
    return ruby_pipeline_element(blocks)

def Id():
    return nn.Identity()

class apply_at(nn.Module):
    def __init__(self, N, oper):
        super().__init__()
        self.N = N
        self.oper = oper
        
    def forward(self, x):
        x_prime = self.oper(x[self.N])
        x[self.N] = x_prime
        return x

class fst(apply_at):
    def __init__(self, oper):
        super().__init__(0, oper)

class snd(apply_at):
    def __init__(self, oper):
        super().__init__(1, oper)

class lst(nn.Module):
    def __init__(self, N=None):
        super().__init__()
        self.N = N

    def forward(self, x):
        if self.N is None:
            return x[-1]
        return x[-self.N:]

class pick_at(nn.Module):
    def __init__(self, N):
        super().__init__()
        self.N = N

    def forward(self, x):
        return x[self.N]

class drop_at(nn.Module):
    def __init__(self, N):
        super().__init__()
        self.N = N

    def forward(self, x):
        begin = x[:self.N]
        tail = x[self.N + 1:]
        return torch.cat((begin, tail), dim=0)

class fork_inv(nn.Module):
    def __init__(self, N):
        super().__init__()
        self.N = N

    def forward(self, x):
        return x[::-1]

def shape_transform_1d(from_shape, to_shape):
    return nn.Linear(from_shape, to_shape)

def act(act):
    match act:
        case "relu":
            return nn.ReLU()
        case "sigmoid":
            return nn.Sigmoid()
        case "tanh":
            return nn.Tanh()
        case _:
            raise ValueError(f"Invalid activation function: {act}")

def activation(act):
    return act(act)

def repeat_block(block, n, same_block=True, block_fn=True, block_args=None, block_kwargs=None):
    if block_fn:
        if not same_block:
            return nn.Sequential(*[block(*block_args, **block_kwargs) for _ in range(n)])
        else:
            block = block(*block_args, **block_kwargs)
    return nn.Sequential(*[block for _ in range(n)])

def relu():
    return nn.ReLU()

def sigmoid():
    return nn.Sigmoid()

def tanh():
    return nn.Tanh()

def softmax():
    return nn.Softmax()

def gelu():
    return nn.GELU()

def elu():
    return nn.ELU()

def selu():
    return nn.SELU()

def silu():
    return nn.SiLU()

def mish():
    return nn.Mish()



import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def plot_training_loss(losses, epochs=None, title="Training and Validation Loss", 
                       figsize=(10, 6), save_path=None, show=True):
    train_losses = losses['train_loss']
    val_losses = losses['val_loss']
    
    if epochs is None:
        epochs = np.arange(1, len(train_losses) + 1)
    
    data = pd.DataFrame({
        'Epoch': np.concatenate([epochs, epochs]),
        'Loss': np.concatenate([train_losses, val_losses]),
        'Type': ['Training'] * len(train_losses) + ['Validation'] * len(val_losses)
    })
    
    plt.figure(figsize=figsize)
    sns.set_style("whitegrid")
    
    ax = sns.lineplot(x='Epoch', y='Loss', hue='Type', data=data, marker='o')
    
    plt.title(title, fontsize=15)
    plt.xlabel('Epoch', fontsize=12)
    plt.ylabel('Loss', fontsize=12)
    
    plt.tight_layout()
    
    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
    if show:
        plt.show()
    
    return plt.gcf()


import torch
import torch.nn as nn

def _loss_fn(loss):
    if isinstance(loss, str):
        return _str_to_loss(loss)
    elif isinstance(loss, nn.Module):
        return loss
    else:
        raise ValueError(f"Unsupported loss function: {loss}")


def _str_to_loss(loss):
    loss = loss.lower()
    match loss:
        case 'huber':
            return torch.nn.HuberLoss()
        case 'l1' | 'mae':
            return torch.nn.L1Loss()
        case 'l2' | 'mse':
            return torch.nn.MSELoss()
        case 'smooth_l1':
            return torch.nn.SmoothL1Loss()
        case 'soft_margin':
            return torch.nn.SoftMarginLoss()
        case 'nll':
            return torch.nn.NLLLoss()
        case 'cross_entropy':
            return torch.nn.CrossEntropyLoss()
        case 'binary_cross_entropy', 'bce':
            return torch.nn.BCELoss()
        case 'binary_cross_entropy_with_logits', 'bce_logits':
            return torch.nn.BCEWithLogitsLoss()
        case 'poisson_nll':
            return torch.nn.PoissonNLLLoss()
        case 'kl_div':
            return torch.nn.KLDivLoss()
        case 'cosine_similarity':
            return torch.nn.CosineSimilarity()
        case 'cosine_embedding_loss':
            return torch.nn.CosineEmbeddingLoss()
        case 'hinge_embedding_loss':
            return torch.nn.HingeEmbeddingLoss()
        case 'multi_margin_loss':
            return torch.nn.MultiMarginLoss()
        case 'multi_label_margin_loss':
            return torch.nn.MultiLabelMarginLoss()
        case 'multi_label_soft_margin_loss':
            return torch.nn.MultiLabelSoftMarginLoss()
        case 'multi_label_cross_entropy':
            return torch.nn.MultiLabelCrossEntropyLoss()
        case 'triplet_margin_loss':
            return torch.nn.TripletMarginLoss()
        case 'triplet_margin_with_distance_loss':
            return torch.nn.TripletMarginWithDistanceLoss()
        case 'cosine_embedding_loss':
            return torch.nn.CosineEmbeddingLoss()
        case 'margin_ranking_loss':
            return torch.nn.MarginRankingLoss()
        case 'hinge_loss':
            return torch.nn.HingeLoss()
        case 'cosine_similarity':
            return torch.nn.CosineSimilarity()
        case 'cosine_embedding_loss':
            return torch.nn.CosineEmbeddingLoss()
        case 'margin_ranking_loss':
            return torch.nn.MarginRankingLoss()

    raise ValueError(f"Unsupported loss function: {loss}")



from setuptools import setup, find_packages

setup(
    name='torch-gamma',
    version='0.0.4',
    packages=find_packages(),
    install_requires=[
        'torch',
        'numpy',
        'scikit-learn',
        'matplotlib',
        'seaborn'
    ],
    author='KevinZonda',
    author_email='realkevin@tutanota.com',
    description='PyTorch γ is a high level library inspired by Ruby',
    keywords="pytorch, machine learning",
    url='https://github.com/KevinZonda/pytorch-gamma'
)